@@ -1,4 +1,4 @@
-#ifndef NDARRAY_BASIC_ARRAY
+﻿#ifndef NDARRAY_BASIC_ARRAY
 #define NDARRAY_BASIC_ARRAY
 
 #include <memory>
@@ -12,6 +12,7 @@
 #include <algorithm>
 #include <atomic>
 
+#include <librapid/math/rapid_math.hpp>
 #include <librapid/ndarray/to_string.hpp>
 
 // Define this if using a custom cblas interface.
@@ -21,6 +22,8 @@
 #include <cblas.h>
 #endif
 
+#include <librapid/ndarray/cblas_api.hpp>
+
 namespace librapid
 {
 	namespace ndarray
@@ -54,6 +57,21 @@ namespace librapid
 			nd_allocator<typename std::common_type<A_T, B_T>::type>>
 			operator/(const A_T &val, const basic_ndarray<B_T, B_A> &arr);
 
+		/**
+		 * \rst
+		 * 
+		 * The core multi-dimensional array class that is responsible for handling
+		 * all array functionality.
+		 * 
+		 * It is highly optimized, fully templated, and is capable of using CBlas
+		 * interfaces to accelerate many of its functions.
+		 * 
+		 * Arrays of up to 32 dimensions can be created, though this can be
+		 * overridden by editing the "config.hpp" file, or by defining the macro
+		 * ``ND_MAX_DIMS`` before including ``librapid.hpp``.
+		 * 
+		 * \endrst
+		 */
 		template<typename T, class alloc, typename std::enable_if<std::is_arithmetic<T>::value, int>::type>
 		class basic_ndarray
 		{
@@ -82,39 +100,32 @@ namespace librapid
 		public:
 
 			/**
-			 * This is a function. It does something.
-			 *
-			 * Hello, World!
-			 *
 			 * \rst
-			 *
-			 * Hello. This is a title
-			 * ======================
-			 *
-			 * 1. This
-			 * 2. Is
-			 * 3. A
-			 * 4. Numbered
-			 * 5. List
-			 *
-			 *
-			 * .. code-block:: Python
-			 *
-			 * 		print("Hello, World!")
-			 *
-			 * .. hint::
-			 * 		This is a hint in a box! How cool is that!?
-			 *
+			 * 
+			 * Create an empty n-dimensional array. This array does not have an
+			 * extent or a stride, and many functions will not operate correctly
+			 * on such an array.
+			 * 
+			 * .. Hint::
+			 *		No memory is allocated on the heap when using this function,
+			 *		so it is incredibly fast.
+			 * 
 			 * \endrst
-			 *
 			 */
-			void set_stride(const stride &s)
-			{
-				m_stride = s;
-			}
-
 			basic_ndarray() = default;
 
+			/**
+			 * \rst
+			 * 
+			 * Create a new array from a given extent.
+			 * 
+			 * The array created will have the same number of dimensions
+			 * as the number of elements passed in the extent object. For
+			 * example, passing in ``extent(2, 3)`` will create a 2x3
+			 * matrix.
+			 * 
+			 * \endrst
+			 */
 			template<typename V>
 			basic_ndarray(const basic_extent<V> &size) : m_extent(size),
 				m_stride(stride::from_extent(size.get_extent(), size.ndim())),
@@ -220,6 +231,11 @@ namespace librapid
 				return (nd_int) m_extent.ndim();
 			}
 
+			ND_INLINE nd_int size() const
+			{
+				return (nd_int) m_extent_product;
+			}
+
 			ND_INLINE bool is_initialized() const
 			{
 				return m_origin_references != nullptr;
@@ -502,11 +518,11 @@ namespace librapid
 			}
 
 			template<typename O>
-			ND_INLINE void reshape(const std::vector<O> &order)
+			ND_INLINE void reshape(const basic_extent<O> &new_shape)
 			{
-				if (math::product(order) != m_extent_product)
+				if (math::product(new_shape.get_extent(), new_shape.ndim()) != m_extent_product)
 					throw std::length_error("Array sizes are different, so cannot reshape array. Shapes "
-											+ m_extent.str() + " and " + extent(order).str() + " cannot be broadcast");
+											+ m_extent.str() + " and " + new_shape.str() + " cannot be broadcast");
 
 				if (!m_stride.is_trivial())
 				{
@@ -558,28 +574,41 @@ namespace librapid
 					m_origin_size = m_extent_product;
 				}
 
-				m_stride = stride::from_extent(order);
-				m_extent = extent(order);
+				m_stride = stride::from_extent(std::vector<O>(new_shape.begin(),
+											   new_shape.end()));
+				m_extent = extent(new_shape);
 			}
 
 			template<typename O>
-			ND_INLINE void reshape(const std::initializer_list<O> &order)
+			ND_INLINE void reshape(const std::initializer_list<O> &new_shape)
 			{
-				reshape(std::vector<O>(order.begin(), order.end()));
+				reshape(std::vector<O>(new_shape.begin(), new_shape.end()));
 			}
 
 			template<typename O>
-			ND_INLINE basic_ndarray<T, alloc> reshaped(const std::vector<O> &order) const
+			ND_INLINE void reshape(const std::vector<O> &new_shape)
+			{
+				reshape(std::vector<O>(new_shape.begin(), new_shape.end()));
+			}
+
+			template<typename O>
+			ND_INLINE basic_ndarray<T, alloc> reshaped(const basic_extent<O> &new_shape) const
 			{
 				auto res = create_reference();
-				res.reshape(order);
+				res.reshape(new_shape);
 				return res;
 			}
 
 			template<typename O>
-			ND_INLINE basic_ndarray<T, alloc> reshaped(const std::initializer_list<O> &order)
+			ND_INLINE basic_ndarray<T, alloc> reshaped(const std::initializer_list<O> &new_shape) const
 			{
-				return reshaped(std::vector<O>(order.begin(), order.end()));
+				return reshaped(std::vector<O>(new_shape.begin(), new_shape.end()));
+			}
+
+			template<typename O>
+			ND_INLINE basic_ndarray<T, alloc> reshaped(const std::vector<O> &new_shape) const
+			{
+				return reshaped(std::vector<O>(new_shape.begin(), new_shape.end()));
 			}
 
 			ND_INLINE void strip_front()
@@ -747,10 +776,19 @@ namespace librapid
 			{
 				using R_T = typename std::common_type<T, B_T>::type;
 
+				bool is_matrix_vector = false;
+
 				const auto &o_e = other.get_extent();
-				if (utils::check_ptr_match(o_e.get_extent(), o_e.ndim(),
-					utils::sub_vector(m_extent.get_extent(), m_extent.ndim(), 1),
-					true))
+				bool is_matrix_vector_like = utils::check_ptr_match(o_e.get_extent(), o_e.ndim(),
+																	utils::sub_vector(m_extent.get_extent(),
+																	m_extent.ndim(), 1), true);
+
+				// Check for column-vector
+				if (ndim() == 2 && other.ndim() == 2)
+					if (m_extent[1] == o_e[0] && o_e[1] == 1)
+						is_matrix_vector = true;
+
+				if (is_matrix_vector || is_matrix_vector_like)
 				{
 					// Matrix-Vector product
 					nd_int res_shape[ND_MAX_DIMS]{};
@@ -761,15 +799,39 @@ namespace librapid
 						res_shape[i] = other.get_extent()[i];
 
 					auto res = basic_ndarray<R_T>(extent(res_shape, dims));
-					for (nd_int i = 0; i < m_extent[0]; i++)
-						res[i] = subscript(i).dot(other);
+
+					// #ifdef LIBRAPID_CBLAS
+					const auto M = m_extent[0];
+					const auto N = m_extent[1];
+					const auto K = other.get_extent()[0];
+
+					if (is_matrix_vector_like)
+					{
+						for (nd_int i = 0; i < M; i++)
+							res[i] = subscript(i).dot(other);
+						return res;
+					}
+
+					const R_T alpha = 1.0;
+					const R_T beta = 0.0;
+
+					const auto trans = m_stride[0] != N;
+
+					const auto lda = M;
+					const auto ldb = other.get_stride()[other.ndim() - 1];
+
+					auto *__restrict a = m_data_start;
+					auto *__restrict b = other.get_data_start();
+					auto *__restrict c = res.get_data_start();
+
+					linalg::cblas_gemv('r', trans, M, N, alpha, a, lda, b, ldb, beta, c, 1);
 
 					return res;
 				}
 
 				if (ndim() != other.ndim())
-					throw std::range_error("Cannot compute dot product on arrays with " +
-										   m_extent.str() + " and " + other.get_extent().str());
+					throw std::domain_error("Cannot compute dot product on arrays with " +
+											m_extent.str() + " and " + other.get_extent().str());
 
 				nd_int dims = ndim();
 
@@ -778,104 +840,23 @@ namespace librapid
 					case 1:
 						{
 							if (m_extent[0] != other.get_extent()[0])
-								throw std::range_error("Cannot compute dot product with arrays with " +
-													   m_extent.str() + " and " + other.get_extent().str());
+								throw std::domain_error("Cannot compute dot product with arrays with " +
+														m_extent.str() + " and " + other.get_extent().str());
 
 							// Vector product
 							basic_ndarray<R_T, nd_allocator<R_T>> res(extent({1}));
 							res.m_is_scalar = true;
 
-						#ifndef LIBRAPID_CBLAS
-							R_T *r = res.get_data_start();
-							T *m = m_data_start;
-							B_T *o = other.get_data_start();
-							nd_int lda = m_stride[0];
-							nd_int ldb = other.get_stride()[0];
-							*r = 0;
-							for (nd_int i = 0; i < m_extent_product; i++)
-								*r += m[i * lda] * o[i * ldb];
-						#else
-							*res.get_data_start() = cblas_ddot(m_extent_product, m_data_start, m_stride[0],
-															   other.get_data_start(), other.get_stride()[0]);
-						#endif // LIBRAPID_CBLAS
+							*res.get_data_start() = linalg::cblas_dot(m_extent_product, m_data_start, m_stride[0],
+																	  other.get_data_start(), other.get_stride()[0]);
 
 							return res;
 						}
 					case 2:
 						{
 							if (m_extent[1] != other.get_extent()[0])
-								throw std::range_error("Cannot compute dot product with arrays with " +
-													   m_extent.str() + " and " + other.get_extent().str());
-
-						#ifndef LIBRAPID_CBLAS
-							const basic_ndarray<T, alloc> tmp_this = *this;
-							const basic_ndarray<T, alloc> tmp_other = other.transposed_matrix();
-
-							const auto M = get_extent()[0];
-							const auto N = get_extent()[1];
-							const auto K = tmp_other.get_extent()[1];
-							const auto K_prime = other.get_extent()[1];
-
-							auto res = basic_ndarray<R_T>(extent{M, K_prime});
-
-							T *a = m_data_start;
-							B_T *b = tmp_other.get_data_start();
-							R_T *c = res.get_data_start();
-
-							nd_int lda = m_stride[0], fda = m_stride[1];
-							nd_int ldb = tmp_other.get_stride()[0], fdb = tmp_other.get_stride()[1];
-							nd_int ldc = res.get_stride()[0], fdc = res.get_stride()[1];
-
-							nd_int index_a, index_b, index_c;
-
-							// Only run in parallel if arrays are smaller than a
-							// given size. Running in parallel on smaller matrices
-							// will result in slower code. Note, a branch is used
-							// in preference to #pragma omp ... if (...) because
-							// that requires runtime evaluation of a condition to
-							// set up threads, which adds a significant overhead
-							if (M * N * K < 25000)
-							{
-								for (nd_int outer = 0; outer < M; ++outer)
-								{
-									for (nd_int inner = 0; inner < K_prime; ++inner)
-									{
-										index_c = outer * ldc + inner * fdc;
-										c[index_c] = 0;
-
-										for (nd_int k = 0; k < N; k++)
-										{
-											index_a = outer * lda + k * fda;
-											index_b = inner * ldb + k * fdb;
-
-											c[index_c] += a[index_a] * b[index_b];
-										}
-									}
-								}
-							}
-							else
-							{
-							#pragma omp parallel for shared(M, N, K, K_prime, a, b, c, lda, ldb, ldc, fda, fdb, fdc) private(index_a, index_b, index_c) default(none) num_threads(ND_NUM_THREADS)
-								for (nd_int outer = 0; outer < M; ++outer)
-								{
-									for (nd_int inner = 0; inner < K_prime; ++inner)
-									{
-										index_c = outer * ldc + inner * fdc;
-										c[index_c] = 0;
-
-										for (nd_int k = 0; k < N; k++)
-										{
-											index_a = outer * lda + k * fda;
-											index_b = inner * ldb + k * fdb;
-
-											c[index_c] += a[index_a] * b[index_b];
-										}
-									}
-								}
-							}
-
-							return res;
-						#else
+								throw std::domain_error("Cannot compute dot product with arrays with " +
+														m_extent.str() + " and " + other.get_extent().str());
 
 							const auto M = m_extent[0];
 							const auto N = m_extent[1];
@@ -886,8 +867,10 @@ namespace librapid
 
 							auto res = basic_ndarray<R_T>(extent{M, K});
 
-							const auto transA = m_stride[0] == N ? CblasNoTrans : CblasTrans;
-							const auto transB = other.get_stride()[0] == K ? CblasNoTrans : CblasTrans;
+							// const auto transA = m_stride[0] == N ? CblasNoTrans : CblasTrans;
+							// const auto transB = other.get_stride()[0] == K ? CblasNoTrans : CblasTrans;
+							const auto transA = m_stride[0] != N;
+							const auto transB = other.get_stride()[0] != K;
 
 							const auto lda = N;
 							const auto ldb = K;
@@ -897,20 +880,17 @@ namespace librapid
 							auto *__restrict b = other.get_data_start();
 							auto *__restrict c = res.get_data_start();
 
-							cblas_dgemm(CblasRowMajor, transA, transB, M, K, N,
-										alpha, a, lda, b, ldb, beta, c, ldc);
+							linalg::cblas_gemm('r', transA, transB, M, K, N,
+											   alpha, a, lda, b, ldb, beta, c, ldc);
 
 							return res;
-						#endif // LIBRAPID_CBLAS
 						}
 					default:
 						{
 							// Check the arrays are valid
 							if (m_extent[ndim() - 1] != other.get_extent()[other.ndim() - 2])
-							{
-								throw std::range_error("Cannot compute dot product with arrays with " +
-													   m_extent.str() + " and " + other.get_extent().str());
-							}
+								throw std::domain_error("Cannot compute dot product with arrays with " +
+														m_extent.str() + " and " + other.get_extent().str());
 
 							// Create the new array dimensions
 							nd_int new_extent[ND_MAX_DIMS]{};
@@ -1451,7 +1431,8 @@ namespace librapid
 			nd_allocator<typename std::common_type<A_T, B_T>::type>>
 			operator+(const A_T &val, const basic_ndarray<B_T, B_A> &arr)
 		{
-			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr, []<typename T_a, typename T_b>(T_a a, T_b b)
+			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr,
+																	[]<typename T_a, typename T_b>(T_a a, T_b b)
 			{
 				return a + b;
 			});
@@ -1462,7 +1443,8 @@ namespace librapid
 			nd_allocator<typename std::common_type<A_T, B_T>::type>>
 			operator-(const A_T &val, const basic_ndarray<B_T, B_A> &arr)
 		{
-			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr, []<typename T_a, typename T_b>(T_a a, T_b b)
+			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr,
+																	[]<typename T_a, typename T_b>(T_a a, T_b b)
 			{
 				return a - b;
 			});
@@ -1473,7 +1455,8 @@ namespace librapid
 			nd_allocator<typename std::common_type<A_T, B_T>::type>>
 			operator*(const A_T &val, const basic_ndarray<B_T, B_A> &arr)
 		{
-			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr, []<typename T_a, typename T_b>(T_a a, T_b b)
+			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr,
+																	[]<typename T_a, typename T_b>(T_a a, T_b b)
 			{
 				return a * b;
 			});
@@ -1484,12 +1467,689 @@ namespace librapid
 			nd_allocator<typename std::common_type<A_T, B_T>::type>>
 			operator/(const A_T &val, const basic_ndarray<B_T, B_A> &arr)
 		{
-			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr, []<typename T_a, typename T_b>(T_a a, T_b b)
+			return basic_ndarray<B_T, B_A>::scalar_array_arithmetic(val, arr,
+																	[]<typename T_a, typename T_b>(T_a a, T_b b)
 			{
 				return a / b;
 			});
 		}
 
+		/**
+		 *\rst
+		 *
+		 *.. Hint::
+		 *	This function is mostly for compatibility
+		 *	with the C# port of the library, as the
+		 *	C++ and Python libraries support overloaded
+		 *	operators.
+		 *
+		 *Add two values together and return the result.
+		 *
+		 *The input values can be any type that supports
+		 *addition. In general, the return type will be
+		 *the higher precision of the two input types,
+		 *or an n-dimensional array if one is passed.
+		 *
+		 *Parameters
+		 *----------
+		 *
+		 *addend1: any
+		 *	The left-hand side of the addition operation
+		 *addend2: any
+		 *	The right-hand side of the addition operation
+		 *
+		 *Returns
+		 *-------
+		 *
+		 *sum: any
+		 *	The result of the addition calculation
+		 *
+		 *\endrst
+		 */
+		template<typename T_A, typename T_B>
+		ND_INLINE auto add(const T_A &addend1, const T_B &addend2)
+		{
+			return addend1 + addend2;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * .. Hint::
+		 *		This function is mostly for compatibility
+		 *		with the C# port of the library, as the
+		 *		C++ and Python libraries support overloaded
+		 *		operators.
+		 *
+		 * Subtract one value from another and return the
+		 * result.
+		 *
+		 * The input values can be any type that supports
+		 * subtraction. In general, the return type will be
+		 * the higher precision of the two input types,
+		 * or an n-dimensional array if one is passed.
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * minuend: any
+		 *		The left-hand side of the subtraction operation
+		 * subtrahend: any
+		 *		The right-hand side of the subtraction operation
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * difference: any
+		 *		The result of the subtraction calculation
+		 *
+		 * \endrst
+		 */
+		template<typename T_A, typename T_B>
+		ND_INLINE auto sub(const T_A &minuend, const T_B &subtrahend)
+		{
+			return minuend - subtrahend;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * .. Hint::
+		 *		This function is mostly for compatibility
+		 *		with the C# port of the library, as the
+		 *		C++ and Python libraries support overloaded
+		 *		operators.
+		 *
+		 * Multiply two values together and return the result
+		 *
+		 * The input values can be any type that supports
+		 * multiplication. In general, the return type will be
+		 * the higher precision of the two input types,
+		 * or an n-dimensional array if one is passed.
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * factor1: any
+		 *		The left-hand side of the multiplication operation
+		 * factor2: any
+		 *		The right-hand side of the multiplication operation
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * product: any
+		 *		The result of the multiplication calculation
+		 *
+		 * \endrst
+		 */
+		template<typename T_A, typename T_B>
+		ND_INLINE auto mul(const T_A &factor1, const T_B &factor2)
+		{
+			return factor1 * factor2;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * .. Hint::
+		 *		This function is mostly for compatibility
+		 *		with the C# port of the library, as the
+		 *		C++ and Python libraries support overloaded
+		 *		operators.
+		 *
+		 * Divide one value by another and return the result.
+		 *
+		 * The input values can be any type that supports
+		 * division. In general, the return type will be
+		 * the higher precision of the two input types,
+		 * or an n-dimensional array if one is passed.
+		 *
+		 * .. Attention::
+		 *		This function does not provide any division-
+		 *		by-zero checking, so NaNs may be produced
+		 *		unknowingly
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * dividend: any
+		 *		The left-hand side of the division operation
+		 * divisor: any
+		 *		The right-hand side of the division operation
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * quotient: any
+		 *		The result of the division calculation
+		 *
+		 * \endrst
+		 */
+		template<typename T_A, typename T_B>
+		ND_INLINE auto div(const T_A &dividend, const T_B &divisor)
+		{
+			return dividend / divisor;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the exponent of all the values in an
+		 * array and return a new array with those values.
+		 *
+		 * Example:
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[1, 2]   =>  [[  2.71828   7.38906  20.0855 ]
+		 *		 [3, 4]]       [ 54.5982  148.413   403.429  ]]
+		 *
+		 * The array that is returned from the function will
+		 * be of the same type as the input, and will have the
+		 * same extent.
+		 *
+		 * .. Hint::
+		 *		Although the extent of the returned array is
+		 *		the same as the input array, the stride may be
+		 *		altered to ensure that it's data is contiguous
+		 *		in memory, improving performance dramatically.
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the exponentiation calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> exp(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::exp(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the sine of all the values in an
+		 * array and return a new array with those values.
+		 *
+		 * .. Attention::
+		 *		The sine function expects the values to be
+		 *		in radians rather than degrees, so ensure
+		 *		to convert if this is undesired.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[1.570 3.141 4.712]   =>  [[  1.  0. -1.]
+		 *        [6.282 7.853 9.424]]       [ -0.  1.  0.]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the sine calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> sin(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::sin(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the cosine of all the values in an
+		 * array and return a new array with those values.
+		 *
+		 * .. Attention::
+		 *		The cosine function expects the values to be
+		 *		in radians rather than degrees, so ensure
+		 *		to convert if this is undesired.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[1.570 3.141 4.712]   =>  [[ 0. -1.  0.]
+		 *        [6.282 7.853 9.424]]       [ 1.  0. -1.]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the cosine calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> cos(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::cos(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the tangent of all the values in an
+		 * array and return a new array with those values.
+		 *
+		 * .. Attention::
+		 *		The tangent function expects the values to be
+		 *		in radians rather than degrees, so ensure
+		 *		to convert if this is undesired.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[1.570 3.141 4.712]   =>  [[ NaN  0.  NaN]
+		 *        [6.282 7.853 9.424]]       [ 0.  NaN   0.]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the tangent calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> tan(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::tan(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the inverse sine of all the values in an
+		 * array and return a new array with those values.
+		 *
+		 * .. Attention::
+		 *		The inverse sine function returns values in
+		 *		radians, not degrees, so ensure this is
+		 *		taken into account when using this function.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[ 0.   1.   0. ]  => [[ 0.        1.5708    0.      ]
+		 *		 [ 0.5  0.  -0.5]]     [ 0.523599  0.       -0.523599]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the inverse sine calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> asin(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::asin(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the inverse cosine of all the values in
+		 * an array and return a new array with those values.
+		 *
+		 * .. Attention::
+		 *		The inverse cosine function will return values
+		 *		in radians, no degrees, so ensure this is taken
+		 *		into account.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[ 0.   1.   0. ]  => [[1.5708 0.     1.5708]
+		 *		 [ 0.5  0.  -0.5]]     [1.0472 1.5708 2.0944]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the inverse cosine calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> acos(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::acos(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the inverse tangent of all the values in
+		 * an array and return a new array with those values.
+		 *
+		 * .. Attention::
+		 *		The inverse tangent function will return values
+		 *		in radians, no degrees, so ensure this is taken
+		 *		into account.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[ 0.   1.   0. ]  => [[ 0.        0.785398  0.      ]
+		 *		 [ 0.5  0.  -0.5]]     [ 0.463648  0.       -0.463648]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the inverse tangent calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> atan(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::atan(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the hyperbolic sine of each element in an
+		 * array, and return a new array containing those values
+		 *
+		 * .. Attention::
+		 *		This function deals in radians, not degrees,
+		 *		so ensure to convert your array before using
+		 *		this function if it will lead to unwanted
+		 *		results.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[-1.  -0.5  0. ]   =>  [[-1.1752   -0.521095  0.      ]
+		 *		 [ 0.   0.5  1. ]]       [ 0.        0.521095  1.1752  ]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the hyperbolic sin function calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> sinh(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::sinh(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the hyperbolic cosine of each element in an
+		 * array, and return a new array containing those values
+		 *
+		 * .. Attention::
+		 *		This function deals in radians, not degrees,
+		 *		so ensure to convert your array before using
+		 *		this function if it will lead to unwanted
+		 *		results.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[-1.  -0.5  0. ]   =>  [[1.54308 1.12763 1.     ]
+		 *		 [ 0.   0.5  1. ]]       [1.      1.12763 1.54308]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the hyperbolic cosine function calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> cosh(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::cosh(value);
+			});
+
+			return res;
+		}
+
+		/**
+		 * \rst
+		 *
+		 * Calculate the hyperbolic tangent of each element in an
+		 * array, and return a new array containing those values
+		 *
+		 * .. Attention::
+		 *		This function deals in radians, not degrees,
+		 *		so ensure to convert your array before using
+		 *		this function if it will lead to unwanted
+		 *		results.
+		 *
+		 * .. Hint::
+		 *		In the future, there will be a function to
+		 *		automatically convert between angle units,
+		 *		so the issue above will be much easier to
+		 *		avoid.
+		 *
+		 * Example (approximate values shown):
+		 *
+		 * .. code-block:: Python
+		 *
+		 *		[[-1.  -0.5  0. ]   =>  [[-0.761594 -0.462117  0.      ]
+		 *		 [ 0.   0.5  1. ]]       [ 0.        0.462117  0.761594]]
+		 *
+		 * Parameters
+		 * ----------
+		 *
+		 * None
+		 *
+		 * Returns
+		 * -------
+		 *
+		 * result: basic_ndarray<T, A>
+		 *		The result of the hyperbolic tangent function calculation
+		 *
+		 * \endrst
+		 */
+		template<typename A_T, typename A_A>
+		ND_INLINE basic_ndarray<A_T, A_A> tanh(const basic_ndarray<A_T, A_A> &arr)
+		{
+			auto res = arr.clone();
+			arithmetic::array_op(res.get_data_start(), arr.get_data_start(),
+								 arr.get_extent(), res.get_stride(), arr.get_stride(),
+								 [](A_T value)
+			{
+				return std::tanh(value);
+			});
+
+			return res;
+		}
+
+		template<typename T, class alloc, typename O = nd_int>
+		ND_INLINE basic_ndarray<T, alloc> reshape(const basic_ndarray<T, alloc> &arr, const basic_extent<O> &new_shape)
+		{
+			return arr.reshaped(new_shape);
+		}
+
 		using ndarray = basic_ndarray<double>;
 		using ndarray_f = basic_ndarray<float>;
 		using ndarray_i = basic_ndarray<int>;
